"""Schema compilation: Finite State Machine implementation
for content models
"""

__version__="$Revision: 1.6 $"
# $Id: FSM.py,v 1.6 2004/04/01 13:31:43 ht Exp $

# FSM for schema
import copy
import types

from Group import All, Sequence, Choice
from Element import Element
from Wildcard import Wildcard
from XSV.infoset.XMLInfoset import Element as ISElement
from XSV.infoset.XMLInfoset import Characters

import Particle

class FSM:
  def __init__(self):
    self.nodes = []
    self.eltTab = {}
    self.startNode = None

  def assignIDs(self):
    if not self.nodes:
      return
    self.nextID = 1
    s = self.startNode
    s.assignIDs()

  def printme(self,file):
    self.assignIDs()
    self.nodes.sort(FSMNode.compareIDs)
    for n in self.nodes:
      if n.isEndNode:
	file.write("*")
      else:
	file.write(" ")
      file.write("%2d:" % n.id)
#      if n.label:
#  	sys.stdout.write("[")
#  	for nn in n.label:
#  	  sys.stdout.write("%d " % nn.id)
#  	sys.stdout.write("]")
      for e in n.edges:
	file.write(" %s->%d" % (e.label[0] or e.label[1], e.dest.id))
      file.write("\n")

  def asXML(self):
    res=ISElement(None,None,"fsm")
    self.assignIDs()
    self.nodes.sort(FSMNode.compareIDs)
    for n in self.nodes:
      ne=res.newDaughter("node")
      ne.newAttr('id',("%s" % n.id))
      if n.isEndNode:
	ne.newAttr('final','true')
      n.edges.sort(FSMEdge.compareLabels)
      for e in n.edges:
        ee=ne.newDaughter("edge")
        if e.label[0] is None:
          lab=unicode(e.label[1].allowed)
        else:
          lab=unicode(e.label[0])
        ee.newAttr('label',lab)
        ee.newAttr('dest',("%s" % e.dest.id))
    return res

  # cf Aho & Ullman p93

  def determinise(self):
    D = FSM()
    D.startNode = FSMNode(D)
    D.startNode.label = _eclosure([self.startNode])
    for nn in D.startNode.label:
      if nn.isEndNode:
	D.startNode.isEndNode = 1
	break
    D.unmarkedNodes = [D.startNode]
    while D.unmarkedNodes:
      x = D.unmarkedNodes[0]
      del D.unmarkedNodes[0]
      destnodes = {}
      for n in x.label:
	for e in n.edges:
	  if destnodes.has_key(e.label):
	    if not e.dest in destnodes[e.label]:
	      destnodes[e.label].append(e.dest)
	  elif e.label is not None:
	    destnodes[e.label] = [e.dest]
      for (label,T) in destnodes.items():
	y = _eclosure(T)
	found = 0
	for n in D.nodes:
	  if y == n.label:
	    for e in x.edges:
	      if e.label == label and e.dest is n:
		break
	    else:
	      FSMEdge(label, x, n)
	    break
	else:
	  n = FSMNode(D)
	  D.unmarkedNodes.append(n)
	  n.label = y
	  for nn in y:
	    if nn.isEndNode:
	      n.isEndNode = 1
	      break
	  FSMEdge(label, x, n)
    return D

  def subsumed(self, fsm):
    # check whether self is subsumed by fsm, by parsing all paths thru self
    unprocStates = [(fsm.startNode,self.startNode)]
    procStates = []
    if self.startNode.isEndNode:
      dEndStates={self.startNode:fsm.startNode.isEndNode or fsm.startNode}
    else:
      dEndStates={}
    while unprocStates:
      p=unprocStates[0]
      b=p[0]
      d=p[1]
      for de in d.edges:
        win=0
        for be in b.edges:
          sr=de.subsumed(be)
          if type(sr)==types.StringType:
            return ('notSubType',de,d,b,sr)
          if sr:
            if sr==2:
              # two element edges
              myDec=de.label[1]
              otherDec=be.label[1]
              if (myDec is not None and
                  otherDec is not None and
                  (type(myDec.scope) is not types.StringType or
                   type(otherDec.scope) is not types.StringType) and
                  not myDec.typeDefinition.isSubtype(otherDec.typeDefinition,['extension']+otherDec.typeDefinition.final)):
                return ('notSubType',de,d,b,be)
            win=1
            # check end status
            dd=de.dest
            bd=be.dest
            if dd.isEndNode:
              if not dEndStates.has_key(dd):
                dEndStates[dd]=bd.isEndNode or bd
              elif bd.isEndNode:
                dEndStates[dd]=1
            ns=(bd,dd)
            if ns not in procStates:
              if ns not in unprocStates:
                unprocStates.append(ns)
            break
        if not win:
          return ('noparse',de,d,b)
      procStates.append(p)
      unprocStates=unprocStates[1:]
    for (d,ok) in dEndStates.items():
      if ok!=1:
        return ('notEnd',d,ok)

  def subsumptionError(self,badGuy,other,checkResult,variety,rName,bName):
    if len(checkResult)==3:
      (rType,nodeR,nodeB) = checkResult
    elif len(checkResult)==4:
      (rType,edgeR,nodeR,nodeB) = checkResult
    else:
      (rType,edgeR,nodeR,nodeB,extra) = checkResult
      
    pics=[Characters(None,"\nBase model:\n"),
          other.asXML(),
          Characters(None,"\nThis model:\n"),
          self.asXML()]
    if rType=='noparse':
      lab=edgeR.label[0]
      badGuy.error("Content model of %s not actually a restriction of base %s %s: a %s element at node %s of this model can't be matched by any edge leaving node %s of the base"%(rName,variety,bName,lab,nodeR.id,nodeB.id),0,pics)
    elif rType=='notSubType':
      lab=edgeR.label[0]
      dtd=edgeR.label[1].typeDefinition.name
      if type(extra)==types.StringType:
        badGuy.error("Content model of %s not actually a restriction of base %s %s: the %s element at node %s of this model has a type (%s) not derived from the type (%s) of the global element accepted by the matching wildcard edge leaving node %s of the base"%(rName,variety,bName,lab,nodeR.id,dtd,extra,nodeB.id),0,pics)
      else:
        btd=extra.label[1].typeDefinition.name
        badGuy.error("Content model of %s not actually a restriction of base %s %s: the %s element at node %s of this model has a type (%s) not derived from the type (%s) of the matching %s edge leaving node %s of the base"%(rName,variety,bName,lab,nodeR.id,dtd,btd,lab,nodeB.id),0,pics)
    elif rType=='notEnd':
      badGuy.error("Content model of %s not actually a restriction of base %s %s: no way forward from this models' end state %s from the base's node %s which gets to an end state"%(rName,variety,bName,nodeR.id,nodeB.id),0,pics)
    else:
      shouldnt('notSubsumed')
    

class FSMEdge:
  def __init__(self, label, source, dest):
    # label is (qname,decl), if qname is None, decl is a Wildcard
    self.label = label
    self.source = source
    self.dest = dest
    source.edges.append(self)

  def compareLabels(self,other):
    if self.label[0] < other.label[0]:
      return -1
    elif self.label[0] > other.label[0]:
      return +1
    else:
      return 0

  def subsumed(self,other):
    # is this edge subsumed by other
    dq=self.label[0]
    bq=other.label[0]
    if bq is not None:
      if dq is not None and bq == dq:
        return 2
      else:
        return 0
    else:
      bw=other.label[1]                 # element (dq) vs. wildcard (bw)
      if dq is not None:
        if bw.allows(dq.uri or None):
          # Wildcard vis a vis explicit decl -- by cases wrt processContents
          if bw.processContents=="skip":
            return 1
          else:
            # I don't _think_ it makes sense to tryHardForDecl here -- there
            # might not _be_ any instance
            if (bw.schema.vElementTable.has_key(dq) and
                bw.schema.vElementTable[dq]):
              gd=bw.schema.vElementTable[dq]
              if (self.label is not None and
                  self.label[1].typeDefinition.isSubtype(gd.typeDefinition,['extension']+gd.typeDefinition.final)):
                return 1
              else:
                return gd.typeDefinition.name
            else:
              # lose if strict, as base wouldn't allow elt (no global defn)
              return bw.processContents=="lax"
        else:
          return 0
      else:
        # both wildcards
        if self.label[1].subsumed(bw):
          return 1
        else:
          return 0

class FSMNode:
  id=None
  def __init__(self, fsm):
    fsm.nodes.append(self)
    self.fsm = fsm
    self.isEndNode = 0
    self.edges = []
    self.label = None
    self.mark = 0

  def assignIDs(self):
    if self.id:
      return
#    print "assigning id %d to %s" % (self.fsm.nextID,self)
    self.id = self.fsm.nextID
    self.fsm.nextID = self.fsm.nextID + 1
    for e in self.edges:
      e.dest.assignIDs()

  def compareIDs(self,other):
    if self.id < other.id:
      return -1
    elif self.id > other.id:
      return +1
    else:
      return 0

# cf Aho & Ullman p92

def _eclosure(T):
  for n in T:
    n.mark=1
  STACK = copy.copy(T)
  ECLOSURE = copy.copy(T)
  while STACK:
    s = STACK[0]
    del STACK[0]
    for e in s.edges:
      if e.label == None:
	t = e.dest
	if not t.mark:
          t.mark=1
	  ECLOSURE.append(t)
	  STACK.insert(0, t)
  # ECLOSURE.sort()			# so we can compare them easily???
  for n in ECLOSURE:
    n.mark=0
  return ECLOSURE

# Change labels of form (name, place) to name

def _relabelFSM(fsm):
  for n in fsm.nodes:
    for e in n.edges:
      if type(e.label)  is  types.TupleType:
        e.label = e.label[0]

# Check a FSM is deterministic

def _checkFSM(fsm):
  for n in fsm.nodes:
    for e in range(len(n.edges)):
      l = n.edges[e].label
      lq=l[0]
      for f in range(0,e):
        m = n.edges[f].label
        mq=m[0]
        if lq is None:
          # wildcard
          lw=l[1]
          if mq is None:
            # wildcard
            mw=m[1]
            if not lw.intersect(mw).isEmpty():
              return "%s/%s" % (lw,mw)
          else:
            if lw.allows(mq.uri):
              return "%s/%s" % (lw,mq)
        else:
          if mq is None:
            # wildcard
            mw=m[1]
            if mw.allows(lq.uri):
              return "%s/%s" % (lq,mw)
          else:
            if lq == mq:
              return "%s/%s" % (lq,mq)
  return 0

def _translateA(self, next, place):
  # doesn't enforce cardinality yet
  n = FSMNode(next.fsm)
  i = 0
  for particle in self.particles:
    m = particle.translate(n, (place,i))
    i = i+1
    FSMEdge(None, n, m)
  FSMEdge(None, n, next)
  return n

def _translateE(self, next, place):
  # what do I do if there are no edges, i.e. abstract element with
  # no descendants?
  fsm = next.fsm
  eltTab = fsm.eltTab
  n = FSMNode(fsm)
  if self.abstract!='true':
    FSMEdge(((self.qname,self),place), n, next)
  if eltTab.has_key(self.qname):
    if self.typeDefinition  is not  eltTab[self.qname].typeDefinition:
      self.error("illegal redeclaration of %s" % self.qname)
  else:
    eltTab[self.qname]=self
  if (type(self.scope) is types.StringType and
      'substitution' not in self.prohibitedSubstitutions):
    # used to be self.schema.vElementTable[qname].equivClass -- why???
    for e in self.equivClass:
      if e.abstract!='true':
        FSMEdge(((e.qname,e),place), n, next)
      if eltTab.has_key(e.qname):
        if e.typeDefinition  is not  eltTab[e.qname].typeDefinition:
          self.error("illegal redeclaration of %s" % e.qname)
      else:
        eltTab[e.qname]=e
  return n

def _translateW(self, next, place):
  n = FSMNode(next.fsm)
  FSMEdge(((None, self),place), n, next)
  return n

def _translateC(self, next, place):
  n = FSMNode(next.fsm)
  i = 0
  for particle in self.particles:
    m = particle.translate(next, (place,i))
    i = i+1
    FSMEdge(None, n, m)
  return n

def _translateS(self, next, place):
  n = next
  rmodel = copy.copy(self.particles)
  rmodel.reverse()
  i = 0
  for particle in rmodel:
    n = particle.translate(n, (place,i))
    i = i+1
  return n

def _translateP(self, next, place):
  fsm = next.fsm
  n = next
  if self.term is None:
     if self.termType=='element' and self.termName:
       # we can check the content model, recursive type check may fail
       qq=self.termName
     else:
       qq=Particle.UndefQName
  if self.occurs[1] and self.occurs[1]>101:
    self.schema.error('performance restriction means max>101 means unbounded',
                      self.xrpr.elt,1)
    self.occurs=(self.occurs[0],None)
  if not self.occurs[1]:
    t = FSMNode(fsm)
    if self.term is not None:
      n = self.term.translate(t, place)
    else:
      n = FSMNode(n.fsm)
      FSMEdge(((qq,None),place),n,t)
    FSMEdge(None, t, n)
    FSMEdge(None, n, next)
  else:
    for i in range(self.occurs[0], self.occurs[1]):
      if self.term is not None:
        n = self.term.translate(n, place)
      else:
        m=n
        n = FSMNode(n.fsm)
        FSMEdge(((qq,None),place),n,m)
      FSMEdge(None, n, next)
  if self.occurs[0]>101:
    self.schema.error('performance restrictions means min>101 means 101',
                      self.xrpr.elt,1)
    self.occurs=(101,self.occurs[1])
  for i in range(0, self.occurs[0]): 
    if self.term is not None:
      n = self.term.translate(n, place)
    else:
      m=n
      n = FSMNode(n.fsm)
      FSMEdge(((qq,None),place),n,m)
  return n

def UniqueFSM(particle):
  ndfsm = FSM()
  end = FSMNode(ndfsm)
  end.isEndNode = 1
  start = particle.translate(end, ())
  ndfsm.startNode = start
  fsm = ndfsm.determinise()
  _relabelFSM(fsm)
  return (fsm,_checkFSM(fsm))

def init():
  Particle.Particle.translate=_translateP
  Sequence.translate=_translateS
  Choice.translate=_translateC
  Wildcard.translate=_translateW
  Element.translate=_translateE
  All.translate=_translateA





# $Log: FSM.py,v $
# Revision 1.6  2004/04/01 13:31:43  ht
# work on final/block a bit
#
# Revision 1.5  2003/12/04 10:52:42  ht
# handle display of subsumption failures locally
#
# Revision 1.4  2002/11/25 11:39:35  ht
# protect against missing decl
#
# Revision 1.3  2002/11/11 18:18:40  ht
# use unicode properly for names/dumping
#
# Revision 1.2  2002/11/05 14:18:22  ht
# package FSM construction and checking and move to FSM
#
# Revision 1.1  2002/06/28 09:40:22  ht
# XSV as package: components
#
# Revision 1.2  2002/06/10 14:38:53  ht
# fix uba
#
# Revision 1.1  2002/05/24 22:33:14  ht
# split out of XMLSchema
#
