"""Schema compilation: ListFacet component"""

__version__="$Revision: 1.7 $"
# $Id: ListFacet.py,v 1.7 2003/06/13 10:50:12 ht Exp $

import re

from Facet import Facet

class ListFacet(Facet):
  def register(self,table):
    if table.has_key(self.name):
      table[self.name].stringValue.append(self.stringValue)
    else:
      table[self.name]=self
      self.stringValue=[self.stringValue]

class Pattern(ListFacet):
  name='pattern'
  def __getattr__(self,name):
    if name=='regexps':
      if self.value is None:
        self.regexps=None
      else:
        self.regexps=map(lambda pat:_tryCompile(pat),
                         self.value)
      return self.regexps
    elif name=='value':
      return self.stringValue           # don't cache
    else:
      raise AttributeError,name

class Enumeration(ListFacet):
  name='enumeration'
  def __getattr__(self,name):
    if name=="value":
      self.value=map(lambda sv,s=self:s.getValue(sv),
                     self.stringValue)
      self.value=filter(lambda x:x is not None,
                        self.value)
      return self.value
    else:
      raise AttributeError,name

  def getValue(self,string):
    if 'whiteSpace' in self.type.facets:
      ff={'whiteSpace':self.type.facets['whiteSpace']}
    else:
      ff={}
    res=self.type.validateText(string,self,self,ff,1)
    if res:
      self.error("facet enumeration value %s not a valid literal: %s"%(string,
                                                                       res))
    else:
      return self.actualValue

def init():
  global _badPat,_trivPat
  # note we allow the classes d, D, s, S, w and W because the Python
  # semantics are very nearly correct
  # Note in reading this that \\ puts a \ in the string, thereby escaping
  # the following char for the regexp, i.e \\\\ puts a \ in the regexp
  _badPat=re.compile("(^|[^\\\\])\\\\[^nrt\\|.?*\\\\+(){}\\[\\]\\-\\^dDsSwW]|(\\[[^\[\]]*\\]-\\[)",re.UNICODE)
  _trivPat=re.compile('.*',re.DOTALL)              # match anything, sigh

def _tryCompile(pat):
  if _badPat.search(pat) is not None:
    return _trivPat
  try:
    return re.compile("(%s)$"%pat)      # force match of whole input
  except re.error:
    return _trivPat

# $Log: ListFacet.py,v $
# Revision 1.7  2003/06/13 10:50:12  ht
# make patterns match whole input
#
# Revision 1.6  2003/03/03 20:55:39  ht
# allow \\\\ in pattern string we can compile
#
# Revision 1.5  2002/11/25 14:56:34  ht
# fix rebuilding of Enumerations to compute value
#
# Revision 1.4  2002/11/25 10:39:01  ht
# pervasive changes to shift to using values for simple types where required by REC
#
# Revision 1.3  2002/09/24 14:10:27  ht
# try allowing some charclasses (dsw) which Python supports well enough
#
# Revision 1.2  2002/09/23 21:33:19  ht
# compile patterns as Python regexps on demand,
# falling back to no check (via .*) if non-Python re features used
#
# Revision 1.1  2002/06/28 09:40:22  ht
# XSV as package: components
#
