"""Distutils setup for XSV
   $Id: setup.py,v 1.8 2004/01/31 15:24:39 ht Exp $"""

import sys

from distutils.core import setup
from distutils.command.install_data import install_data

#data installer with improved intelligence over distutils.
#data files are copied into the project directory instead
#of willy-nilly
class smart_install_data(install_data):
  def run(self):
    #need to change self.install_dir to the actual library dir
    install_cmd = self.get_finalized_command('install')
    self.install_dir = getattr(install_cmd, 'install_lib')
    return install_data.run(self)

datafiles=[('XSV/example',
            ['triv.xml',
             'triv.xsd',
             'tiny.xml',
             'tiny.xsd']),
           ('XSV',['XSV/nndump.xml',
                   'XSV/XMLSchema.dtd',
                   'XSV/datatypes.dtd']),
           ('XSV/pubtext',['xsv.xsl']),
           ('XSV/doc',['COPYING',
                       'COPYRIGHT',
                       'pc-shrinkwrap',
                       'xsv-status.xml',
                       'xsv-status.html'])] # only works for bdist and friends

if sys.platform.find("win32",0,5)==0:
  platform="WIN32"
else:
  platform="UN*X"

setup(name='XSV',
      version='2.7',
      description='W3C XML Schema compiler and validator',
      author='Henry S. Thompson',
      author_email='HThompson@ed.ac.uk',
      url="http://www.ltg.ed.ac.uk/~ht/xsv-status.html",
      license="GPL",
      long_description="""Implements the bulk of the W3C XML Schema Recommendation,
      including support for XML representation of schemas and schema-validity
      assessment of instances.""",
      platforms=[platform],
      packages=['XSV','XSV.compile','XSV.compile.elts',
                'XSV.validate','XSV.infoset','XSV.infoset.altNorm',
                'XSV.infoset.relNorm','XSV.infoset.indNorm','XSV.util'],
      cmdclass = {'install_data': smart_install_data},
      data_files=datafiles)
